// Read/Write request processors for fileio driver
// Generated by Walt Oney's driver wizard

#include "stddcls.h"
#include "driver.h"

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchCreate(PDEVICE_OBJECT fdo, PIRP Irp)
	{							// DispatchCreate
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);

	// Disallow attempt to open with additional name qualification

	if (stack->FileObject->FileName.Length)
		{						// extra name qualification given
		KdPrint((DRIVERNAME " - Failing IRP_MJ_CREATE because of extra name component(s) %ws\n",
			stack->FileObject->FileName.Buffer));
		return CompleteRequest(Irp, STATUS_ACCESS_DENIED, 0);
		}						// extra name qualification given

	// Also disallow "relative" open requests

	if (stack->FileObject->RelatedFileObject)
		{						// relative open
		KdPrint((DRIVERNAME " - Failing relative IRP_MJ_CREATE\n"));
		return CompleteRequest(Irp, STATUS_ACCESS_DENIED, 0);
		}						// relative open
	
	// Claim the remove lock to prevent this device object from disappearing while
	// the handle is open

	NTSTATUS status = IoAcquireRemoveLock(&pdx->RemoveLock, stack->FileObject);
	if (NT_SUCCESS(status))
		InterlockedIncrement(&pdx->handles);
	return CompleteRequest(Irp, status, 0);
	}							// DispatchCreate

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchClose(PDEVICE_OBJECT fdo, PIRP Irp)
	{							// DispatchClose
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	InterlockedDecrement(&pdx->handles);
	
	// Release the remove lock to match the acquisition done in DispatchCreate

	IoReleaseRemoveLock(&pdx->RemoveLock, stack->FileObject);
	return CompleteRequest(Irp, STATUS_SUCCESS, 0);
	}							// DispatchClose

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS StartDevice(PDEVICE_OBJECT fdo, PCM_PARTIAL_RESOURCE_LIST raw, PCM_PARTIAL_RESOURCE_LIST translated)
	{							// StartDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	NTSTATUS status;

	// A likely time to want to perform file I/O operations is during configuration of
	// the device, in order to read a firmware file for downloading to the device. This
	// stage is problematic in Win98 because the init order for CONFIGMG is before that
	// of IFSMgr. The portable file I/O library solves that particular problem.

	HANDLE hfile;
	status = OpenFile(L"\\SystemRoot\\System32\\Drivers\\fileio.jnk", TRUE, &hfile);
	if (!NT_SUCCESS(status))
		{
		KdPrint((DRIVERNAME " - OpenFile failed - %X\n", status));
		return status;
		}

	PVOID buffer = NULL;

	__try
		{
		ULONG size = (ULONG) GetFileSize(hfile);
		if (!size)
			__leave;			// file is empty

		buffer = ExAllocatePool(PagedPool, size);
		if (!buffer)
			{
			KdPrint((DRIVERNAME " - Unable to allocate %d bytes for data buffer\n", size));
			status = STATUS_INSUFFICIENT_RESOURCES;
			__leave;
			}
		
		ULONG numread;
		status = ReadFile(hfile, buffer, size, &numread);

		if (!NT_SUCCESS(status))
			{
			KdPrint((DRIVERNAME " - ReadFile failed - %X\n", status));
			__leave;
			}

		KdPrint((DRIVERNAME " - ReadFile succeeded and read %d bytes\n", numread));

		// Preserve the data buffer for the associated test program

		pdx->RandomJunk = buffer;
		pdx->RandomJunkSize = numread;
		buffer = NULL;
		}
	__finally
		{
		CloseFile(hfile);
		if (buffer)
			ExFreePool(buffer);
		}

	return status;
	}							// StartDevice

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

VOID StopDevice(IN PDEVICE_OBJECT fdo, BOOLEAN oktouch /* = FALSE */)
	{							// StopDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	if (pdx->RandomJunk)
		ExFreePool(pdx->RandomJunk);
	pdx->RandomJunk = NULL;
	}							// StopDevice
